/*
   cfgfile.c - handle configuration files

   Copyright (C) 2002 Arthur de Jong.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <syslog.h>

#include "cfgfile.h"
#include "xmalloc.h"
#ifdef HAVE_SETRLIMIT
#include "reslimit.h"
#endif /* HAVE_SETRLIMIT */


/* TODO: make this something better (maybe  a struct) */
extern char *cvsd_rootjail;
extern int cvsd_nice;
extern uid_t cvsd_uid;
extern gid_t cvsd_gid;
extern char *cvsd_pidfile;
extern int cvsd_port;
extern int cvsd_maxconnections;
extern char *cvs_cmd;
extern mode_t cvs_umask;
void add_cvsarg(char *arg);


/* maximum linelength in the configuration file */
#define MAX_LINE_LENGTH 1024


/* find out the uid this string describes
   returns 1 on success, 0 on failure */
static int get_uid(char *str,uid_t *val)
{
  struct passwd *pwent;
  char *tmp;
  /* check if it is a valid numerical uid */
  *val=(uid_t)strtol(str,&tmp,0);
  if ((*str!='\0')&&(*tmp=='\0'))
    return 1;
  /* find by name */
  pwent=getpwnam(str);
  if (pwent!=NULL)
  {
    *val=pwent->pw_uid;
    return 1;
  }
  return 0;
}


/* find the group id this string represents
   returns 1 on success, 0 on failure */
static int get_gid(char *str,gid_t *val)
{
  struct group *grent;
  char *tmp;
  /* check if it is a valid numerical uid */
  *val=(gid_t)strtol(str,&tmp,0);
  if ((*str!='\0')&&(*tmp=='\0'))
    return 1;
  /* find by name */
  grent=getgrnam(str);
  if (grent!=NULL)
  {
    *val=grent->gr_gid;
    return 1;
  }
  return 0;
}


/* read the configuration file */
void read_configfile(char *filename)
{
  FILE *fp;
  int lnr=0;
  char line[MAX_LINE_LENGTH];
  char *tmp;  /* for manipulating the read line */
  char *tmp2; /* for building cvs command args */
  int i;
  int numrepos=0;
#ifdef HAVE_SETRLIMIT
  char *tmp3; /* TODO: really improve this */
  rlim_t resval;
#endif /* HAVE_SETRLIMIT */
  /* open configuration file */
  if ((fp=fopen(filename,"r"))==NULL)
  {
    syslog(LOG_ERR,"cannot open config file (%s): %s",filename,strerror(errno));
    exit(1);
  }
#ifdef DEBUG
  syslog(LOG_DEBUG,"debug: reading config file (%s)",filename);
#endif /* DEBUG */
  /* read all lines in the file */
  while (fgets(line,MAX_LINE_LENGTH,fp)!=NULL)
  {
    lnr++;
    /* strip newline */
    i=(int)strlen(line);
    if ((i>0)&&(line[i-1]=='\n'))
    {
      line[i-1]='\0';
    }
    else
    {
      syslog(LOG_ERR,"%s:%d: line too long or last line missing newline",filename,lnr);
      exit(1);
    }
    /* strip out comments */
    tmp=strchr(line,'#');
    if (tmp!=NULL) *tmp='\0';
    /* convert tabs to spaces */
    tmp=line;
    while (*tmp!='\0')
    {
      if (*tmp=='\t') *tmp=' ';
      tmp++;
    }
    /* remove trailing spaces */
    tmp--;
    while ((tmp>=line)&&(*tmp==' '))
    {
      *tmp='\0';
      tmp--;
    }
    /* skip beginning spaces */
    tmp=line;
    while (*tmp==' ')
      tmp++;
    /* find identifier */
    if (strncmp(tmp,"RootJail ",9)==0)
    {
      /* skip space */
      for (tmp+=9;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: RootJail: argument missing",filename,lnr);
        exit(1);
      }
      /* only accept absolute paths. abort otherwise.*/
      if ((*tmp!='/')&&(strcmp(tmp,"none")!=0))
      {
        syslog(LOG_ERR,"%s:%d: RootJail: '%s' must be an absolute path or 'none'",filename,lnr,tmp);
        exit(1);
      }
      /* save value */
      if (cvsd_rootjail!=NULL)
      {
        syslog(LOG_ERR,"%s:%d: RootJail: can only be specified once",filename,lnr);
      }
      else
      {
        cvsd_rootjail=(char *)xmalloc((strlen(tmp)+1)*sizeof(char));
        strcpy(cvsd_rootjail,tmp);
      }
    }
    else if (strncmp(tmp,"Nice ",5)==0)
    {
      /* skip space */
      for (tmp+=5;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Nice: argument missing",filename,lnr);
        exit(1);
      }
      /* check if it is a valid numerical uid */
      cvsd_nice=(int)strtol(tmp,&tmp2,0);
      if (*tmp2!='\0')
      {
        syslog(LOG_ERR,"%s:%d: Nice: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Uid ",4)==0)
    {
      /* skip space */
      for (tmp+=4;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Uid: argument missing",filename,lnr);
        exit(1);
      }
      if (!get_uid(tmp,&cvsd_uid))
      {
        syslog(LOG_ERR,"%s:%d: Uid: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Gid ",4)==0)
    {
      /* skip space */
      for (tmp+=4;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Gid: argument missing",filename,lnr);
        exit(1);
      }
      if (!get_gid(tmp,&cvsd_gid))
      {
        syslog(LOG_ERR,"%s:%d: Gid: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Port ",5)==0)
    {
      /* skip space */
      for (tmp+=5;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Port: argument missing",filename,lnr);
        exit(1);
      }
      cvsd_port=(int)strtol(tmp,&tmp2,0);
      if (*tmp2!='\0')
      {
        syslog(LOG_ERR,"%s:%d: Port: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"MaxConnections ",15)==0)
    {
      /* skip space */
      for (tmp+=15;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: MaxConnections: argument missing",filename,lnr);
        exit(1);
      }
      cvsd_maxconnections=(int)strtol(tmp,&tmp,0);
      if (*tmp!='\0')
      {
        syslog(LOG_ERR,"%s:%d: MaxConnections: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Repos ",6)==0)
    {
      /* skip space */
      for (tmp+=6;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Repos: argument missing",filename,lnr);
        exit(1);
      }
      /* only accept absolute paths. abort otherwise.*/
      if (*tmp!='/')
      {
        syslog(LOG_ERR,"%s:%d: Repos: '%s' must be an absolute path",filename,lnr,tmp);
        exit(1);
      }
      /* add --allow-root=path to cvs parameters */
      tmp2=(char *)xmalloc((13+strlen(tmp)+1)*sizeof(char));
      strcpy(tmp2,"--allow-root=");
      strcat(tmp2,tmp);
      add_cvsarg(tmp2);
      numrepos++;
    }
    else if (strncmp(tmp,"PidFile ",8)==0)
    {
      /* skip space */
      for (tmp+=8;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: PidFile: argument missing",filename,lnr);
        exit(1);
      }
      if (cvsd_pidfile==NULL)
      {
        cvsd_pidfile=(char *)xmalloc((strlen(tmp)+1)*sizeof(char));
        strcpy(cvsd_pidfile,tmp);
      }
    }
    else if (strncmp(tmp,"CvsCommand ",11)==0)
    {
      /* skip space */
      for (tmp+=11;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: CvsCommand: argument missing",filename,lnr);
        exit(1);
      }
      /* only accept absolute paths, abort otherwise */
      if (*tmp!='/')
      {
        syslog(LOG_ERR,"%s:%d: CvsCommand: '%s' must be an absolute path",filename,lnr,tmp);
        exit(1);
      }
      /* save value */
      if (cvs_cmd!=NULL)
      {
        syslog(LOG_ERR,"%s:%d: CvsCommand: can only be specified once",filename,lnr);
      }
      else
      {
        cvs_cmd=(char *)xmalloc((strlen(tmp)+1)*sizeof(char));
        strcpy(cvs_cmd,tmp);
      }
    }
    else if (strncmp(tmp,"Limit ",6)==0)
    {
#ifdef HAVE_SETRLIMIT
      /* skip space */
      for (tmp+=6;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Limit: resourcename missing",filename,lnr);
        exit(1);
      }
      /* find resource name */
      tmp3=tmp;
      for (;(*tmp!=' ')&&(*tmp!='\0');tmp++);
      for (;*tmp==' ';tmp++) *tmp='\0';
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Limit: limit value missing",filename,lnr);
        exit(1);
      }
      /* read value */
      if ( (strcmp(tmp,"none")==0) ||
           (strcmp(tmp,"inf")==0) ||
           (strcmp(tmp,"unlimited")==0) )
        resval=RLIM_INFINITY;
      else
        resval=0;
      switch (getresourcetype(tmp3))
      {
      case RESTYPE_NUM:  /* numeric value */
      case RESTYPE_SIZE: /* size value */
        if (resval==RLIM_INFINITY) break;
        resval=(rlim_t)strtol(tmp,&tmp2,0);
        if ((tmp2[0]=='k')&&(tmp2[1]=='\0'))
          resval*=(rlim_t)1024;
        else if ((tmp2[0]=='m')&&(tmp2[1]=='\0'))
          resval*=(rlim_t)(1024*1024);
        else if ((tmp2[0]!='\0')&&(tmp2[0]!='b'))
        {
          syslog(LOG_ERR,"%s:%d: Limit: wrong value '%s'",filename,lnr,tmp);
          exit(1);
        }
        break;
      case RESTYPE_TIME: /* time value */
        if (resval==RLIM_INFINITY) break;
        resval=(rlim_t)strtol(tmp,&tmp2,0);
        if ((tmp2[0]=='m')&&(tmp2[1]=='\0'))
          resval*=(rlim_t)60;
        else if ((tmp2[0]==':')&&(tmp2[1]!='\0'))
        {
          tmp2++;
          resval*=(rlim_t)60;
          resval+=(rlim_t)strtol(tmp2,&tmp2,0);
          if (tmp2[0]!='\0')
          {
            syslog(LOG_ERR,"%s:%d: Limit: wrong value '%s'",filename,lnr,tmp);
            exit(1);
          }
        }
        else if ((tmp2[0]!='\0')&&(tmp2[0]!='s'))
        {
          syslog(LOG_ERR,"%s:%d: Limit: wrong value '%s'",filename,lnr,tmp);
          exit(1);
        }
        break;
      default:
        syslog(LOG_ERR,"%s:%d: Limit: unknown or unsupported resource '%s'",filename,lnr,tmp2);
        exit(1);
      }
      /* save the limit, execute with setreslimits() */
      savereslimit(tmp3,resval,resval);
#else /* HAVE_SETRLIMIT */
      syslog(LOG_ERR,"%s:%d: Limit: not supported on this system",filename,lnr);
      exit(1);
#endif /* not HAVE_SETRLIMIT */
    }
    else if (strncmp(tmp,"Umask ",6)==0)
    {
      /* skip space */
      for (tmp+=6;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Umask: argument missing",filename,lnr);
        exit(1);
      }
      /* check if it is a valid numerical uid */
      cvs_umask=(mode_t)strtol(tmp,&tmp2,8);
      if ((*tmp2!='\0')||(cvs_umask<0000)||(cvs_umask>0777))
      {
        syslog(LOG_ERR,"%s:%d: Umask: wrong argument '%s'",filename,lnr,tmp);
        exit(1);
      }
    }
    else if (*tmp!='\0') /* line should be empty */
    {
      syslog(LOG_ERR,"%s:%d: unrecognized line",filename,lnr);
      exit(1);
    }
  }
  fclose(fp);
  if (numrepos==0)
  {
    syslog(LOG_ERR,"%s: contains no 'Repos' statements",filename);
    exit(1);
  }
#ifdef DEBUG
  syslog(LOG_DEBUG,"debug: done reading config file");
#endif /* DEBUG */
}
